/* Copyright (C) 2010-2012 Red Hat, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include <config.h>

#include <alloca.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "logging.h"
#include "iwh.h"

#define PROGRAM_NAME "iwhd"

static enum log_level loglevel;
static const char *logfile;
static FILE *logstream;
static int logcheck_flag;
static pthread_mutex_t logcheck_lock = PTHREAD_MUTEX_INITIALIZER;

static void assure_newline(char *s)
{
	size_t len = strlen(s);
	if (len && s[len-1] == '\n')
		len -= 1;
	s[len] = '\n';
	s[len+1] = 0;
}

static void log_str(const char *fmt, va_list args)
	__attribute__ ((format (printf, 1, 0)));
static void log_str(const char *fmt, va_list args)
{
	char tbuf[sizeof("Jan 01 00:00:00")];
	struct tm tmb;
	size_t len;

	pid_t pid = getpid() & 0xFFFFFFFF;
	len = sizeof("Jan 01 00:00:00 " PROGRAM_NAME "[123456790]: ");
	len += strlen(fmt) + 3;
	char *f = alloca(len);

	time_t now = time(NULL);
	if (localtime_r(&now, &tmb) == NULL ||
	    strftime(tbuf, sizeof("Jan 01 00:00:00"), "%b %d %T", &tmb) == 0)
	{
		strcpy(tbuf, "Unk 00 00:00:00");
	}

	sprintf(f, "%s " PROGRAM_NAME "[%u]: %s", tbuf, pid, fmt);
	assure_newline(f);		/* length accounted above */

	vfprintf(logstream?:stderr, f, args);	/* atomic write */
	/* no fflush, setbuf-ed to _IONBF */
}

void log_msg(enum log_level level, const char *fmt, ...)
{
	va_list args;

	if (level > loglevel)
		return;

	va_start(args, fmt);
	log_str(fmt, args);
	va_end(args);
}

/*
 * TBD: MHD_OPTION_EXTERNAL_LOGGER for MHD_start_daemon. Needs an extern.
 * void log_mhd(void * arg, const char * fmt, va_list ap) { .... }
 * Does Microhttpd ever print anything though? Never seen a message from it.
 */

/*
 * TBD: Once all of our sites are converted, replace with a simpler
 * version that just does freopen(logfile, "a+", logstream, ....).
 */
static int
log_reset(void)
{
	int fd1;

	if (!logfile)
		return 0;

	fd1 = open(logfile, O_WRONLY|O_CREAT|O_APPEND, 0666);	/* use umask */
	if (fd1 == -1) {
		error(0,errno,_("unable to open log `%s'"),logfile);
		return -1;
	}

	/*
	 * Sadly freopen() is unusable if we want two streams pointing
	 * to the same file (not even with mode "a"). We could supply
	 * "/dev/fd/%d" as a filename, but this is too yucky (what if
	 * /dev/fd is not mounted?). So, we substitute the fds 1 and 2 quetly.
	 */
	if (fflush(stdout) != 0) {
		error(0,errno,_("failed to flush stdout"));
		close(fd1);
		return -1;
	}
	if (fflush(stderr) != 0) {
		error(0,errno,_("failed to flush stdout"));
		close(fd1);
		return -1;
	}
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	if (dup2(fd1, 1) == -1) {
		error(0,errno,_("unable to dup log to stdout"));
		close(fd1);
		return -1;
	}
	if (dup2(fd1, 2) == -1) {
		error(0,errno,_("unable to dup log to stderr"));
		close(fd1);
		return -1;
	}
	close(fd1);

	logstream = stderr;
	return 0;
}

static void hup_handler(int arg)
{
	pthread_mutex_lock(&logcheck_lock);
	logcheck_flag = 1;
	pthread_mutex_unlock(&logcheck_lock);
}

void log_check(void)
{
	pthread_mutex_lock(&logcheck_lock);
	if (logcheck_flag) {
		(void) log_reset();
		logcheck_flag = 0;
	}
	pthread_mutex_unlock(&logcheck_lock);
}

void log_init(enum log_level level, const char *path)
{
	struct sigaction act;

	loglevel = level;

	if (!path)
		return;

	logfile = path;
	if (log_reset() != 0)
		exit(EXIT_FAILURE);	/* always have error reported already */

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler = hup_handler;
	act.sa_flags = SA_RESTART;
	sigaction(SIGHUP, &act, NULL);
}
