/* Copyright (C) 2010-2012 Red Hat, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include <config.h>

#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <iostream>
#include "iwh.h"
#include "logging.h"
#include "meta.h"
#include "query.h"

using namespace std;

/* Mongo (rather antisocially) tries to define this itself. */
#undef VERSION

#include <mongo/client/dbclient.h>
using namespace mongo;

/* TBD: parameterize */
#define MAIN_TBL "repo.main"

/*
 * Since the client isn't inherently MT-safe, we serialize access to it
 * ourselves.  Fortunately, none of our metadata operations should be very
 * long-lived; if they are it probably means our connection is FUBAR and other
 * threads will be affected anyway.
 */

#define SHOW_CONTENTION

static pthread_mutex_t client_lock = PTHREAD_MUTEX_INITIALIZER;

#ifdef SHOW_CONTENTION
#define CLIENT_LOCK do {					\
	if (pthread_mutex_trylock(&client_lock) != 0) {		\
		log_msg(IW_LOG_INFO,"contention in %s",__func__);	\
		pthread_mutex_lock(&client_lock);		\
	}							\
} while (0)
#else
#define CLIENT_LOCK	pthread_mutex_lock(&client_lock)
#endif
#define CLIENT_UNLOCK	pthread_mutex_unlock(&client_lock)

void
dbl_to_str (double *foo, char *optr)
{
	unsigned int i;
	unsigned char *iptr = (unsigned char *)foo;

	for (i = 0; i < sizeof(*foo); ++i) {
		optr += sprintf(optr,"%02x",*(iptr++));
	}
}

class RepoMeta;
class RepoQuery;

class RepoMeta {
	DBClientConnection	client;
	char			addr[128];

public:
		RepoMeta	();
		~RepoMeta	();

	char *	DidPut		(const char *bucket, const char *key,
				 const char *loc, size_t size);
	void	GotCopy		(const char *bucket, const char *key,
				 const char *loc);
	char *	HasCopy		(const char *bucket, const char *key,
				 const char *loc);
	int	SetValue	(const char *bucket, const char *key,
				 const char *mkey, const char * mvalue);
	int	GetValue	(const char *bucket, const char *key,
				 const char *mkey, char ** mvalue);
	RepoQuery * NewQuery	(const char *bucket, const char *key,
				 const char * expr);
	auto_ptr<DBClientCursor> GetCursor (Query &q);
	void	Delete		(const char *bucket, const char *key);
	size_t	GetSize		(const char *bucket, const char *key);
	int	Check		(const char *bucket, const char *key,
				 const char *depot);
	void *	GetAttrList	(const char *bucket, const char *key);
};

class RepoQuery {
	RepoMeta &		parent;
	DBClientCursor *	curs;
	value_t *		expr;
public:
		RepoQuery	(const char *, const char *, const char *,
				 RepoMeta &);
		~RepoQuery	();
	bool	Next		(void);
	bool	Dead		(void);
	char	*bucket;
	char	*key;
	getter_t getter;
};

static RepoMeta *it;

RepoMeta::RepoMeta () :
    client(true)
{
	if (!verbose) {
		cout.rdbuf(0);
		cout << "bite me" << endl;
	}

	// TBD: assemble this string properly
	snprintf(addr,128,"%s:%u",db_host,db_port);
	try {
		client.connect(addr);
	}
	catch (ConnectException &ce) {
		log_msg(IW_LOG_INFO, "server down, no metadata access");
	}
}

extern "C" void
meta_init (void)
{
	it = new RepoMeta();
}

RepoMeta::~RepoMeta ()
{
}

extern "C" void
meta_fini (void)
{
	delete it;
}

auto_ptr<DBClientCursor>
RepoMeta::GetCursor (Query &q)
{
	auto_ptr<DBClientCursor> curs;

	for (;;) {
		bool	retry	= false;
		try {
			curs = client.query(MAIN_TBL,q);
		}
		catch (ConnectException &ce) {
			log_msg(IW_LOG_INFO, "reconnection to %s failed", addr);
			curs.reset();
		}
		catch (SocketException &ce) {
			if (ce._type != SocketException::FAILED_STATE) {
				log_msg(IW_LOG_INFO,
					"connection to %s error", addr);
			}
			else {
				/* TBD: this message loops */
				log_msg(IW_LOG_INFO,
					"reconnection to %s error", addr);
				retry = true;
			}
			curs.reset();
		}
		if (!retry)
			break;
		/*
		 * This wait exists to defeat the check for lastReconnectTry in
		 * DBClientConnection::_checkConnection, and no other reason.
		 * If we do actually fail to reconnect, we must report it up.
		 */
		struct timespec ts;
		ts.tv_sec = 3;
		ts.tv_nsec = 0;
		nanosleep(&ts, NULL);
	}
	return curs;
}

char *
RepoMeta::DidPut (const char *bucket, const char *key, const char *loc,
		  size_t size)
{
	BSONObjBuilder			bb;
	struct timeval			now_tv;
	double				now;
	auto_ptr<DBClientCursor>	curs;
	Query				q;
	char				now_str[sizeof(now)*2+1];

	gettimeofday(&now_tv,NULL);
	now = (double)now_tv.tv_sec + (double)now_tv.tv_usec / 1000000.0;
	dbl_to_str(&now,now_str);
	log_msg(IW_LOG_DEBUG,"now_str is %s", now_str);

	q = QUERY("_bucket"<<bucket<<"_key"<<key);
	curs = GetCursor(q);
	if (!curs.get()) {
		log_msg(IW_LOG_INFO,"DidPut failed for %s/%s",bucket,key);
		return NULL;
	}
	if (curs->more()) {
		/* Nice functionality, but what an ugly syntax! */
		client.update(MAIN_TBL,q,BSON(
			"$set"<<BSON("_loc"<<BSON_ARRAY(loc))
		<<	"$set"<<BSON("_date"<<now)
		<<	"$set"<<BSON("_etag"<<now_str)
		<<	"$set"<<BSON("_size"<<(long long)size)));
#if 0
		client.update(MAIN_TBL,q,
			BSON("$set"<<BSON("_loc"<<BSON_ARRAY(loc))));
		client.update(MAIN_TBL,q,
			BSON("$set"<<BSON("_date"<<now)));
		client.update(MAIN_TBL,q,
			BSON("$set"<<BSON("_etag"<<now_str)));
		client.update(MAIN_TBL,q,
			BSON("$set"<<BSON("_size"<<(long long)size)));
#endif
	}
	else {
		bb << "_bucket" << bucket << "_key" << key
		   << "_loc" << BSON_ARRAY(loc) << "_date" << now
		   << "_etag" << now_str << "_size" << (long long)size;
		client.insert(MAIN_TBL,bb.obj());
	}

	return strdup(now_str);
}

extern "C" char *
meta_did_put (const char *bucket, const char *key, const char *loc, size_t size)
{
	char	*rc;

	CLIENT_LOCK;
	rc = it->DidPut(bucket,key,loc,size);
	CLIENT_UNLOCK;

	return rc;
}

void
RepoMeta::GotCopy (const char *bucket, const char *key, const char *loc)
{
	BSONObjBuilder			bb;
	auto_ptr<DBClientCursor>	curs;
	Query				q;

	q = QUERY("_bucket"<<bucket<<"_key"<<key);
	curs = GetCursor(q);
	if (!curs.get()) {
		log_msg(IW_LOG_INFO,"GotCopy failed for %s/%s",bucket,key);
		return;
	}
	if (curs->more()) {
		/* Nice functionality, but what an ugly syntax! */
		client.update(MAIN_TBL,q,BSON("$addToSet"<<BSON("_loc"<<loc)));
	}
	else {
		log_msg(IW_LOG_INFO,"%s/%s not found in GotCopy!",bucket,key);
	}
}

extern "C" void
meta_got_copy (const char *bucket, const char *key, const char *loc)
{
	CLIENT_LOCK;
	it->GotCopy(bucket,key,loc);
	CLIENT_UNLOCK;
}

char *
RepoMeta::HasCopy (const char *bucket, const char *key, const char *loc)
{
	BSONObjBuilder			bb;
	auto_ptr<DBClientCursor>	curs;
	Query				q;
	const char			*value;

	q = QUERY("_bucket"<<bucket<<"_key"<<key<<"_loc"<<loc);
	curs = GetCursor(q);
	if (!curs.get()) {
		log_msg(IW_LOG_INFO,"HasCopy failed for %s/%s", bucket, key);
		return NULL;
	}
	if (!curs->more()) {
		log_msg(IW_LOG_INFO,"%s/%s not found at %s", bucket, key, loc);
		return (char *)"";
	}

	value = curs->next().getStringField("_etag");
	if (!value || !*value) {
		log_msg(IW_LOG_INFO,"%s/%s no _etag at %s", bucket, key, loc);
		return (char *)"";
	}

	log_msg(IW_LOG_DEBUG,"%s/%s _etag = %s", bucket, key, value);
	return strdup(value);
}

extern "C" char *
meta_has_copy (const char *bucket, const char *key, const char *loc)
{
	char	*rc;

	CLIENT_LOCK;
	rc = it->HasCopy(bucket,key,loc);
	CLIENT_UNLOCK;

	return rc;
}

int
RepoMeta::SetValue (const char *bucket, const char *key, const char *mkey,
		    const char * mvalue)
{
	Query	q	= QUERY("_bucket"<<bucket<<"_key"<<key);

	for (;;) {
		bool	retry	= false;
		try {
			client.update(MAIN_TBL,q,BSON("$set"<<BSON(mkey<<mvalue)),1);
		}
		catch (ConnectException &ce) {
			log_msg(IW_LOG_INFO,"SetValue failed for %s/%s:%s",
				bucket,key,mkey);
			return ENOTCONN;
		}
		catch (SocketException &ce) {
			if (ce._type != SocketException::FAILED_STATE) {
				log_msg(IW_LOG_INFO,
					"SetValue error for %s/%s:%s",
					bucket,key,mkey);
				return ENOTCONN;
			}
			log_msg(IW_LOG_INFO, "reconnection to %s failed", addr);
			retry = true;
		}
		if (!retry)
			break;
		struct timespec ts;
		ts.tv_sec = 3;
		ts.tv_nsec = 0;
		nanosleep(&ts, NULL);
	}

	// TBD: check for and propagate errors.
	return 0;
}

extern "C" int
meta_set_value (const char *bucket, const char *key, const char *mkey,
		const char * mvalue)
{
	int	rc;

	CLIENT_LOCK;
	rc = it->SetValue(bucket,key,mkey,mvalue);
	CLIENT_UNLOCK;

	return rc;
}

int
RepoMeta::GetValue (const char *bucket, const char *key, const char *mkey,
		    char ** mvalue)
{
	auto_ptr<DBClientCursor>	curs;
	Query				q;
	BSONObj				bo;
	const char *			data;

	q = QUERY("_bucket"<<bucket<<"_key"<<key);
	curs = GetCursor(q);
	if (!curs.get()) {
		log_msg(IW_LOG_INFO,"GetValue failed for %s/%s:%s",
			bucket,key,mkey);
		return ENOTCONN;
	}
	if (!curs->more()) {
		return ENXIO;
	}

	bo = curs->next();
	data = bo.getStringField(mkey);
	if (!data || !*data) {
		return ENOENT;
	}

	*mvalue = strdup(data);
	if (*mvalue == NULL)
		return ENOMEM;

	return 0;
}

extern "C" int
meta_get_value (const char *bucket, const char *key, const char *mkey,
		char ** mvalue)
{
	int	rc;

	CLIENT_LOCK;
	rc = it->GetValue(bucket,key,mkey,mvalue);
	CLIENT_UNLOCK;

	return rc;
}

RepoQuery::RepoQuery (const char *bucket, const char *key, const char *qstr,
		      RepoMeta &p)
	: parent(p)
{
	Query				q;
	auto_ptr<DBClientCursor>	tmp;

	if (bucket) {
		log_msg(IW_LOG_DEBUG, "bucket is %s", bucket);
		q = QUERY("_bucket"<<bucket);
	}
	else if (key) {
		log_msg(IW_LOG_DEBUG, "key is %s", key);
		q = QUERY("_key"<<key);
	}
	else {
		log_msg(IW_LOG_INFO, "both bucket and key are null for query");
		curs = NULL;
		expr = NULL;
		bucket = NULL;
		key = NULL;
		return;
	}

	/*
	 * TBD: we should really convert our query into one of Mongo's,
	 * and let them do all the work.  Handling the general case
	 * would be pretty messy, but we could handle specific cases
	 * pretty easily.  For example, a very high percentage of
	 * queries are likely to be a single field/value comparison.
	 * For now just punt, but revisit later.
	 */

	if (qstr) {
		expr = parse(qstr);
		if (expr) {
			print_value(expr);
		}
		else {
			log_msg(IW_LOG_INFO, "could not parse %s", qstr);
		}
	}
	else {
		expr = NULL;
	}

	tmp = parent.GetCursor(q);
	if (!tmp.get()) {
		curs = NULL;
	}
	else {
		curs = tmp.release();
	}
	bucket = NULL;
	key = NULL;
}

RepoQuery::~RepoQuery ()
{
	delete curs;
}

extern "C" void
meta_query_stop (void * qobj)
{
	CLIENT_LOCK;
	delete (RepoQuery *)qobj;
	CLIENT_UNLOCK;
}

extern "C" const char *
query_getter (void *ctx, const char *id)
{
	BSONObj *cur_bo = (BSONObj *)ctx;

	return (char *)cur_bo->getStringField(id);
}

bool
RepoQuery::Next (void)
{
	BSONObj		bo;

        if (!curs) {
                return false;
        }

	while (curs->more()) {
		bo = curs->next();
		if (expr) {
			getter.func = query_getter;
			getter.ctx = (void *)&bo;
			if (eval(expr,&getter,NULL) <= 0) {
				continue;
			}
		}
		bucket = (char *)bo.getStringField("_bucket");
		key = (char *)bo.getStringField("_key");
		return true;
	}

	return false;
}

bool
RepoQuery::Dead (void)
{
	return curs == NULL;
}

RepoQuery *
RepoMeta::NewQuery (const char *bucket, const char *key, const char *expr)
{
	RepoQuery	*rc;

	rc = new RepoQuery(bucket,key,expr,*this);
	/*
	 * An exception would be more in C++ spirit, but we're unclear
	 * on the effects of a throw in constructor. So, use Dead() for now.
	 */
	if (rc->Dead()) {
		delete rc;
		return NULL;
	}
	return rc;
}

extern "C" void *
meta_query_new (const char *bucket, const char *key, const char *expr)
{
	void	*rc;

	if ((bucket && key) || (!bucket && !key)) {
		return NULL;
	}

	CLIENT_LOCK;
	rc = it->NewQuery(bucket,key,expr);
	CLIENT_UNLOCK;

	return rc;
}

extern "C" int
meta_query_next (void * qobj, char ** bucket, char ** key)
{
	RepoQuery *	rq	= (RepoQuery *)qobj;

	CLIENT_LOCK;
	if (!rq->Next()) {
		CLIENT_UNLOCK;
		return 0;
	}
	CLIENT_UNLOCK;

	*bucket = rq->bucket;
	*key = rq->key;
	return 1;
}

#if 0
char *
RepoMeta::BucketList (void)
{
	/*
	 * TBD: make this return values instead of producing output.
	 * This is just a code fragment showing how to get a list of buckets,
	 * in case I forget.
	 */
	BSONObj				repl;

	BSONObj dist = BSON("distinct"<<"main"<<"_key"<<"_bucket");
	if (client.runCommand("repo",dist,repl)) {
		cout << repl.toString() << endl;
		BSONObj elem = repl.getField("values").embeddedObject();
		for (int i = 0; i < elem.nFields(); ++i) {
			cout << elem[i].str() << endl;
		}
	}
}
#endif

void
RepoMeta::Delete (const char *bucket, const char *key)
{
	Query	q	= QUERY("_bucket"<<bucket<<"_key"<<key);

	try {
		client.remove(MAIN_TBL,q);
	}
	catch (ConnectException &ce) {
		log_msg(IW_LOG_INFO,"Delete failed for %s/%s",bucket,key);
	}
}

extern "C"
void
meta_delete (const char *bucket, const char *key)
{
	CLIENT_LOCK;
	it->Delete(bucket,key);
	CLIENT_UNLOCK;
}

size_t
RepoMeta::GetSize (const char *bucket, const char *key)
{
	auto_ptr<DBClientCursor>	curs;
	Query				q;
	BSONObj				bo;
	const char *			data;

	(void)data;

	q = QUERY("_bucket"<<bucket<<"_key"<<key);
	curs = GetCursor(q);
	if (!curs.get() || !curs->more()) {
		return 0;
	}

	bo = curs->next();
	return bo.getField("_size").numberLong();
}

extern "C"
size_t
meta_get_size (const char *bucket, const char *key)
{
	size_t	rc;

	CLIENT_LOCK;
	rc = it->GetSize(bucket,key);
	CLIENT_UNLOCK;

	return rc;
}

class AttrList {
public:
				AttrList	(BSONObj &);
	int			Next		(const char **, const char **);
        BSONObj                 obj;
	vector<BSONElement>	vec;
	size_t			idx;
};

AttrList::AttrList (BSONObj &bo)
{
        obj = bo.copy();
	obj.elems(vec);
	idx = 0;
}

int
AttrList::Next (const char **name, const char **value)
{
	BSONElement	elem;

	while (idx < vec.size()) {
		elem = vec[idx];

		// Do not wrap back to 0 when vec.size is SIZE_MAX.
		if (idx != vec.size())
			idx++;

		if (elem.type() == String) {
			*name = elem.fieldName();
			*value = elem.String().c_str();
			return 1;
		}
	}

	return 0;
}

void *
RepoMeta::GetAttrList (const char *bucket, const char *key)
{
	auto_ptr<DBClientCursor>	curs;
	Query				q;
	BSONObj				bo;

	q = QUERY("_bucket"<<bucket<<"_key"<<key);
	curs = GetCursor(q);
	if (!curs.get()) {
		/* This prevents crash, but we return 404 on _attrs. Bad. */
		return NULL;
	}
	if (!curs->more()) {
		return NULL;
	}
	bo = curs->next();

	return new AttrList(bo);
}
extern "C"
void *
meta_get_attrs (const char *bucket, const char *key)
{
	void *poc;

	CLIENT_LOCK;
	poc = it->GetAttrList(bucket,key);
	CLIENT_UNLOCK;

	return poc;
}

extern "C"
int
meta_attr_next (void *ctx, const char **name, const char **value)
{
	AttrList *poc = (AttrList *)ctx;

	return poc->Next(name,value);
}

extern "C"
void
meta_attr_stop (void *ctx)
{
	AttrList *poc = (AttrList *)ctx;

	delete poc;
}
