/* Copyright (C) 2010-2012 Red Hat, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include <config.h>

#include <errno.h>
#include <error.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <jansson.h>

#include "iwh.h"	/* _() */
#include "logging.h"
#include "user.h"
#include "xalloc.h"

static unsigned int userc;
static struct user *userv;
static struct user userU;

struct user *
user_lookup (const char *userid)
{
	unsigned int i;

	if (strcmp(userid, userU.name) == 0)
		return &userU;
	for (i = 0; i < userc; i++) {
		if (strcmp(userid, userv[i].name) == 0) {
			return &userv[i];
		}
	}
	return NULL;
}

void
user_add(const char *usercred)
{
	const char *p = strchr(usercred,':');
	if (!p) {
		error (EXIT_FAILURE, 0,
		       _("The -U argument must contain a colon"));
	}
	if (p == usercred) {
		error (EXIT_FAILURE, 0,
		       _("The user part of -U argument is empty"));
	}
	size_t namelen = p-usercred;
	if (namelen > USERNAME_MAX) {
		error (EXIT_FAILURE, 0, _("User name too long"));
	}
	if (strlen(p+1) > USERPASS_MAX) {
		error (EXIT_FAILURE, 0, _("User password too long"));
	}
	memcpy(userU.name,usercred,namelen);
	userU.name[namelen] = 0;
	strcpy(userU.pass,p+1);
}

static void
user_add_split(unsigned int n, const char *name, const char *secret)
{
	struct user *user = &userv[n];
	if (strlen(name) > USERNAME_MAX) {
		error (EXIT_FAILURE, 0, _("User name too long"));
	}
	if (strlen(secret) > USERPASS_MAX) {
		error (EXIT_FAILURE, 0, _("User password too long"));
	}
	strcpy(user->name,name);
	strcpy(user->pass,secret);
}

static bool
parse_config_inner (json_t *config)
{
	unsigned int	 nusers;
	unsigned int	 i;
	json_t          *userj;
	json_t		*elem;
	const char	*new_key;
	const char	*new_pass;

	if (json_typeof(config) != JSON_ARRAY) {
		log_msg(0,_("user list must be a JSON array"));
		goto err;
	}

	nusers = json_array_size(config);
	if (!nusers) {
		return true;
	}

	userc = nusers;
	userv = xcalloc(userc, sizeof(struct user));

	for (i = 0; i < nusers; ++i) {
		userj = json_array_get(config,i);
		if (!userj) {
			log_msg(0,_("could not get pointer to user %u"),i);
			goto err;
		}

		elem = json_object_get(userj,"name");
		if (!json_is_string(elem)) {
			log_msg(0,_("config elem %u: missing name"),i);
			goto err;
		}
		/* dup_json_string() is not needed */
		new_key = json_string_value(elem);

		/*
		 * Must store unhashed because OAuth requires plaintext. Sad.
		 */
		elem = json_object_get(userj,"secret");
		if (!json_is_string(elem)) {
			log_msg(0,_("config elem %u: missing secret"),i);
			goto err;
		}
		new_pass = json_string_value(elem);

		user_add_split(i, new_key, new_pass);
	}
	return true;

err:
	return false;
}

bool
parse_users (const char *cfg_file)
{
	static json_t	*config;
	json_error_t	 err;
	bool		 ret;

	if (access(cfg_file,R_OK) < 0) {
		error(0,errno,_("failed to open %s for reading"), cfg_file);
		return false;
	}

	config = json_load_file(cfg_file, JANSSON_LOAD_FLAG &err);
	if (!config) {
		log_msg(0,_("JSON error on line %d: %s"),err.line,err.text);
		return false;
	}

	ret = parse_config_inner(config);

	json_decref(config);
	config = NULL;

	return ret;
}
